/*************************************************************************
*                                                                        *
*   Module: Output                                                       *
*                                                                        *
*   Management of options for printing results.                          *
*                                                                        *
*   This file is part of the geneidP Distribution                        *
*                                                                        *
*     Copyright (C) 2000 - Enrique BLANCO GARCIA                         *
*                          Roderic GUIGO SERRA                           *
*                                                                        *
*  This program is free software; you can redistribute it and/or modify  *
*  it under the terms of the GNU General Public License as published by  *
*  the Free Software Foundation; either version 2 of the License, or     *
*  (at your option) any later version.                                   *
*                                                                        *
*  This program is distributed in the hope that it will be useful,       *
*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*  GNU General Public License for more details.                          *
*                                                                        *
*  You should have received a copy of the GNU General Public License     *
*  along with this program; if not, write to the Free Software           *
*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.             *
*************************************************************************/

/* $Id: Output.c,v 1.1 2001/01/08 08:09:48 eblanco Exp $ */

#include "geneid.h"

extern int VRB, FWD, RVS,
           GENAMIC, GENEID,
           GFF;

extern account *m;
        
/* Printing messages (information) */
void printMess(char* s)
{
  if (VRB)
     fprintf (stderr, "> %s\n",s);
}

/* Printing partial and final results (information) */
void printRes(char* s)
{
  if (VRB)
     fprintf (stderr, "\t%s\n",s);
}

/* Printing error messages */
void printError(char *s)
{
     fprintf(stderr,"Error: %s\n",s);
	  exit(1);
}

/* Printing thread messages */
void printThread(char mess[MAXSTRING], int thid)
{
  int i;
  char s[MAXSTRING];

  if (VRB)
    {
      for(i=0; i<thid+1; i++)
	s[i] = '\t';
      s[i] = '\0';

      fprintf(stderr,"%s %s\n",s,mess);
    }
}

void PrintProfile (profile *p, char* signal)
{
  char mess[MAXSTRING];

  sprintf(mess, 
          "Reading... %s:\t%d\t%d\t%d\t(%ld)\t%5.2f\n", 
          signal,
          p->dimension, 
          p->offset, 
          p->order,
          p->dimensionTrans,
          p->cutoff);

  printMess(mess);
}


void OutputHeader()
{
  char* s;
  
  if (GFF)
    printf("## gff-version 2\n");
  
  printf("## source-version: geneidP v 1.0   -- geneid@imim.es\n");
  s = ctime(&m->tStart);

  printf("## date %s",s);
}

/* Print some predictions results according to the options selected */
void POutput(params* p[NUMTHREADS],
	     char* Locus, 
	     char* Sequence,
	     gparam* gp, dict* dAA)
{
  /* Print best gene(s) */
  printMess("Printing Best Gene");
  /* Recover the Best Gene recursively */
  CookingGenes(p[NUMTHREADS-1]->genes->GOptim, Locus, Sequence, gp, dAA);  
}

/* Get time of execution using account information */
void OutputTime()
{
  time_t tEnd;
  int t;
  float caux;
  char mess[MAXSTRING];

  (void) time(&tEnd);
  caux = (float)clock() / (float)CLOCKS_PER_SEC;

  t = (int) tEnd - m->tStart;

  if (t < caux)
    t++;

  /*  sprintf(mess,"CPU time: \t%.3f secs",caux); */
  /*    printRes(mess); */

  sprintf(mess,"Real time: \t%d secs(%.2f mins)\n\n",t,(float)t/MINUTE);
  printRes(mess);
}



