/*************************************************************************
*                                                                        *
*   Module: geneid                                                       *
*                                                                        *
*   Main program. Management of the actions of geneid.                   *
*                                                                        *
*   This file is part of the geneidP Distribution                        *
*                                                                        *
*     Copyright (C) 2000 - Enrique BLANCO GARCIA                         *
*                          Roderic GUIGO SERRA                           *
*     with contributions from:                                           *
*                          Moises  BURSET ALVAREDA                       *
*                          Xavier  MESSEGUER PEYPOCH                     *
*                                                                        *
*  This program is free software; you can redistribute it and/or modify  *
*  it under the terms of the GNU General Public License as published by  *
*  the Free Software Foundation; either version 2 of the License, or     *
*  (at your option) any later version.                                   *
*                                                                        *
*  This program is distributed in the hope that it will be useful,       *
*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*  GNU General Public License for more details.                          *
*                                                                        *
*  You should have received a copy of the GNU General Public License     *
*  along with this program; if not, write to the Free Software           *
*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.             *
*************************************************************************/ 

/* $Id: geneid.c,v 1.2 2001/01/08 08:21:29 eblanco Exp $ */

#include "geneid.h"

/* Setup Flags of GeneId */
int   VRB=0, FWD=1, RVS=1,
      GENAMIC = 1, GENEID = 1, 
      GFF = 0, X10 = 0, 
      EVD = 0, SRP=0;

float EW = NOVALUE; 


/* -- Begin SHARED DATA STRUCTURES -- */
/* Accounting time and results */
account *m;

/* Shared Memory */
char* Sequence;
char* RSequence;

/* Lenght of DNA sequence */
long  LengthSequence;

/* Lenght of DNA sequence of each thread */
long L;

/* Geneid prediction parameters */
gparam** isochores;
int nIsochores;

/* Evidence exons */
packEvidence* evidence;

/* Similarity to protein regions */
packSR* sr;
/* -- End SHARED DATA STRUCTURES -- */


long analizeFile(char* SequenceFile)
{
  struct stat *buf;
  long size;

  buf = (struct stat*)malloc(sizeof(struct stat));

  if (stat(SequenceFile,buf) != 0)
    printError("Impossible to read sequence file");

  size = (long)buf->st_size;
  
  if (size == 0)
    printError("Empty sequence file!");

  free(buf);

  return(size);
}

int SelectIsochore(float percent, gparam** isochores)
{
  int i;
  int stop;

  i = 0;

  percent = PERCENT * percent;

  stop = (percent >= isochores[i]->leftValue && 
          percent <= isochores[i]->rightValue);

  while(!stop)
    {
      i++;
      stop = (percent >= isochores[i]->leftValue && 
              percent <= isochores[i]->rightValue);
    }

  return(i);
}

/*****************************************************************************/

/* Management of splice sites prediction and exon building */
/* LastSequenceThread is the last position of thread sequence */
void  PredictExons(char* Sequence,
		   long LastSequenceThread,
		   packSites* allSites, 
		   packExons* allExons,
		   long l1, long l2,
		   long iRegions[],
		   int Strand, 
		   gparam* gp,
		   int thid)
{
  char mess[MAXSTRING];
  
  long l1a, l1b,
       l2a, l2b,
       l1c, l2c;

  long cutPoint;
  
  /* 0. Define boundaries of splice site prediction
        according to current split positions and strand selected */
  if (Strand == FORWARD)
    {
      /* Forward direction */
      /* Starts and Acc boundaries */
      l1a = l1;
      l2a = (l2 == LastSequenceThread)? l2 : l2 - OVERLAP;

      /* Donors Boundary */
      l1b = l1;
      l2b = l2;

      /* Stops Boundary */
      l1c = l1;
      l2c = l2;

      /* Exons can be built with Stops before cutPoint */
      cutPoint = l1;
    }
  else
    {
      /* Reverse direction */
      /* Starts and Acc Boundaries */
      l1a = l1;
      l2a = l2;

      /* Donor Boundary */
      l1b = (l1 == (LengthSequence-1 - LastSequenceThread))? l1: l1 + OVERLAP;
      l2b = l2;

      /* Stops boundary */
      l1c = l1;
      l2c = l2;

      /* Exons can be built with Stops before cutPoint */
      cutPoint = (l1 == (LengthSequence-1 - LastSequenceThread))? l1 : l1 + OVERLAP;
    }

  /* 1. Predicting splice sites of current split of DNA sequence */ 
  allSites->nStartCodons =
    GetSitesWithProfile(Sequence,gp->StartProfile,allSites->StartCodons,l1a,l2a);

  allSites->nAcceptorSites =
    GetSitesWithProfile(Sequence,gp->AcceptorProfile,allSites->AcceptorSites,l1a,l2a);
  
  allSites->nDonorSites =
    GetSitesWithProfile(Sequence,gp->DonorProfile,allSites->DonorSites,l1b,l2b);

  allSites->nStopCodons =
  GetStopCodons(Sequence,gp->StopProfile, allSites->StopCodons,l1c,l2c);
  
  /* Total number of predicted splice sites in this strand */
  allSites->nSites = allSites->nStartCodons +
      allSites->nAcceptorSites +
      allSites->nDonorSites +
      allSites->nStopCodons;

  sprintf(mess,"Computing sites ... %8ld", allSites->nSites);
  printThread(mess,thid);

  /* 2. Building exons with splice sites predicted before */ 
  allExons->nInitialExons =
    BuildInitialExons(allSites->StartCodons,allSites->nStartCodons,
		      allSites->DonorSites,allSites->nDonorSites,
		      allSites->StopCodons,allSites->nStopCodons,
		      gp->MaxDonors,allExons->InitialExons);

  allExons->nInternalExons =
    BuildInternalExons(allSites->AcceptorSites,allSites->nAcceptorSites,
		       allSites->DonorSites,allSites->nDonorSites,
		       allSites->StopCodons,allSites->nStopCodons,
		       gp->MaxDonors,allExons->InternalExons);

  allExons->nTerminalExons =
    BuildTerminalExons(allSites->AcceptorSites,allSites->nAcceptorSites,
             allSites->StopCodons,allSites->nStopCodons,
             LengthSequence,allExons->TerminalExons,
             cutPoint);

  allExons->nSingles =
    BuildSingles(allSites->StartCodons,allSites->nStartCodons,
		 allSites->StopCodons,allSites->nStopCodons,
		 cutPoint, allExons->Singles);

  /* 3. Scoring and Filtering Exons */
  ScoreExons(Sequence, allExons, l1, l2, Strand, sr, iRegions, isochores, nIsochores); 

  /* Total number of built exons in this strand */
  allExons->nExons = allExons->nInitialExons + 
         allExons->nInternalExons +
         allExons->nTerminalExons +
         allExons->nSingles;

  sprintf(mess,"Computing exons ... %8ld", allExons->nExons);
  printThread(mess,thid);
}

/*****************************************************************************/

void ThreadWork(params* p)
{
  char mess[MAXSTRING];
  long l1,l2;
  int lastSplit;
 
  long currVExons=0;     
  
  /* Measure of C+G content */
  float percent;
  int currentIsochore;
  gparam* gp;

  sprintf(mess,"Starting thread %d(%ld,%ld)",p->quiensoy,p->left,p->right);
  printThread(mess,p->quiensoy);

  /* Split thread sequence in several parts of the same length */
  /* l1 is the current left boundary and l2 is the right */
  l1 = p->left;
  l2 = MIN(p->left + LENGTHSi-1,p->right);
  lastSplit = (l2 == p->right);

  if (EVD)
    InitEvidenceCounter(evidence, l1, p);

  while((l1 < (p->right - OVERLAP))|| (l1 == p->left))
    {
      /* Prediction of sites and exons building */
      
      /* Measure of C+G content of this region */
      percent = MeasureSequence(l1,l2,Sequence);
      sprintf(mess,"C+G-Percent is %f\n", percent);
      printThread(mess,p->quiensoy); 
      
      currentIsochore = SelectIsochore(percent,isochores);
      gp = isochores[currentIsochore];
      sprintf(mess,"Selecting isochore %d\n", currentIsochore+COFFSET);
      printThread(mess,p->quiensoy);
      
      if (FWD) 
	{
	  /* forward sense predictions */ 
	  sprintf(mess,"Thread %d. Running FWD: %ld - %ld",p->quiensoy,l1,l2); 
	  printThread(mess,p->quiensoy); 
	  PredictExons(Sequence,
		       p->right, 
		       p->allSites, 
		       p->allExons, 
		       l1, l2, 
		       p->iRegions,
		       FORWARD,
		       gp,
		       p->quiensoy);
	}      
      if (RVS) 
	{
	  /* reverse sense predictions */ 
	  sprintf(mess,"Thread %d. Running RVS: %ld - %ld",
		  p->quiensoy, LengthSequence-1 - l2, LengthSequence-1 - l1);
	  printThread(mess,p->quiensoy); 
	  PredictExons(RSequence,
		       p->right,
		       p->allSites_r, 
		       p->allExons_r, 
		       LengthSequence-1 - l2, 
		       LengthSequence-1 - l1,
		       p->iRegions,
		       REVERSE,
		       gp,
		       p->quiensoy);
		  
	  /* computing sites positions according to forward sense positions */
	  RecomputePositions(p->allSites_r, LengthSequence);
	  
	  /* exchange acc and donor sites because of Acc < Don needed */
	  SwitchPositions(p->allExons_r);
	}
      
      /* Sort (forward & reverse) exons by acc position */
      if (EVD)
	{
	  /* Searching evidence exons in this split */
	  printThread("Searching evidence exons",p->quiensoy);
	  currVExons = SearchEvidenceExons(evidence,
					   (lastSplit)?l2:l2-OVERLAP,p);
	}

      p->nExons = p->allExons->nExons + p->allExons_r->nExons + currVExons;
      sprintf(mess,"Sorting %ld Exons by Acceptor\n", p->nExons);
      printThread(mess,p->quiensoy);

      /* Merge predicted exons with some evidence exons */
      p->nExons = SortExons(p->allExons, p->allExons_r, 
			    evidence,
			    p->i1vExons, p->i2vExons,
			    p->exons, 
			    l1, l2);
      
      if (EVD)
	SwitchCounters(p);   

      /* Calling genamic for assembling the best gene */ 
      if (GENAMIC && p->nExons)
	{
	  genamic(p->exons, p->nExons, p->genes, gp, p->quiensoy);
	  printThread("Ending GenAmic...",p->quiensoy);
	  
	  /* clean hash table of exons */
	  cleanDumpHash(p->dumpster->h);
	  
	  /* Backup operations of genes to next split */
	  if (!lastSplit)
	    {
	      /* backup of unused genes */
	      printThread("Backup of d-genes",p->quiensoy);
	      BackupArrayD(p->genes, l2 - OVERLAP, gp, p->dumpster);
	      
	      /* back-up best partial genes */
	      printThread("Backup of partial genes",p->quiensoy);
	      BackupGenes(p->genes, gp->nclass, p->dumpster);
	    }
	}
      
      /* Compute boundaries of next split of sequence */
      l1 += LENGTHSi - OVERLAP;
      l2 = MIN(l1 + LENGTHSi -1, p->right);
      lastSplit = (l2 == p->right);
    }/* End of current sequence */

  sprintf(mess,"Ending thread %d",p->quiensoy);
  printThread(mess,p->quiensoy);

  pthread_exit(0);     
}

/*****************************************************************************/

void setThreadParams(params* p[NUMTHREADS],   
		     packSites* allSites[NUMTHREADS],
		     packSites* allSites_r[NUMTHREADS],
		     packExons* allExons[NUMTHREADS],
		     packExons* allExons_r[NUMTHREADS],
		     exonGFF* exons[NUMTHREADS], 
		     long nExons[NUMTHREADS],
		     packGenes* genes[NUMTHREADS],
		     packDump* dumpster[NUMTHREADS])
{
  int i;
  char mess[MAXSTRING];

  sprintf(mess,"GeneID running %d threads",NUMTHREADS);
  printMess(mess);
  sprintf(mess,"LenghtSequence is %ld bp",LengthSequence); 
  printMess(mess);
  
  L = LengthSequence / NUMTHREADS;
  sprintf(mess,"Each thread works with L = %ld bp (not including THREADOVERLAP)", L);
  printMess(mess);

  /* Setting params for threads */
  for(i=0; i<NUMTHREADS; i++)  
    {
      p[i]->quiensoy = i;
      p[i]->left = MAX(0,i * L - THREADOVERLAP);
      p[i]->right = (i<NUMTHREADS-1)?
	(i+1)*L-1 : LengthSequence-1;

      p[i]->allSites = allSites[i];
      p[i]->allSites_r = allSites_r[i];
      p[i]->allExons = allExons[i];
      p[i]->allExons_r = allExons_r[i];
      p[i]->exons = exons[i];
      p[i]->nExons = nExons[i];
      p[i]->genes = genes[i];
      p[i]->dumpster = dumpster[i];

      /* Evidences counters */
      p[i]->i1vExons = 0;
      p[i]->i2vExons = 0;

    }
}

/*****************************************************************************/

void CreateThreads(params* p[NUMTHREADS],
		   pthread_t threads[NUMTHREADS])
{
  int i;
  int res;
  char mess[MAXSTRING];

  printMess("Starting parallelism...");
  for(i=0; i<NUMTHREADS; i++)  
    {
      res = pthread_create(&threads[i],NULL,
			   (void*)ThreadWork,(void*) p[i]);
      if (res != 0)
	{
	  sprintf(mess,"Pthread_create: %s\n",strerror(res));
	  printError(mess);
	}
    }
}

/*****************************************************************************/

void WaitThreads(pthread_t threads[NUMTHREADS])
{
  void* status;
  int i,res;
  char mess[MAXSTRING];
  
  for(i = 0; i< NUMTHREADS; i++)
    {
      res = pthread_join(threads[i],&status);
      if (res != 0)
	{
	  sprintf(mess,"Pthread_join: %s\n",strerror(res));
	  printError(mess);
	}
      if (status != NULL)
	{
	  sprintf(mess,"Wrong finish: Pthread %d !! \n",i);
	  printError(mess);
	}
      else
	{
	  sprintf(mess,"Right finish: Pthread %d\n",i);
	  printMess(mess);
	}
    }
  printMess("Finishing parallelism...");
}

/*****************************************************************************/

int main (int argc, char *argv[])
{
  /* DNA-sequence */
  FILE* seqfile;

  /* Input FileNames */
  char   SequenceFile[FILENAMELENGTH],
         ExonsFile[FILENAMELENGTH],
         SRFile[FILENAMELENGTH],
         ParamFile[FILENAMELENGTH]="";

  /* Forward data structures */
  packSites* allSites[NUMTHREADS];
  packExons* allExons[NUMTHREADS];

  /* Reverse Data Structures */ 
  packSites* allSites_r[NUMTHREADS];
  packExons* allExons_r[NUMTHREADS];

  /* Sort by acceptor Exons */ 
  exonGFF* exons[NUMTHREADS];
  long nExons[NUMTHREADS]; 

  /* Best partial predicted genes */
  packGenes* genes[NUMTHREADS];
  
  /* Dumpster for back-up operations between splits */  
  packDump* dumpster[NUMTHREADS];

  /* Name of current sequence */ 
  char Locus[LOCUSLENGTH];
  char nextLocus[LOCUSLENGTH];
  int  reading;
  
  /* Amino-acid dictionary */
  dict* dAA;

  pthread_t threads[NUMTHREADS];
  params* p[NUMTHREADS];

  char mess[MAXSTRING];     

 
  /* -- initializing stats... -- */
  m = (account*)InitAcc();

  /* 0. Read setup options */
  readargv(argc,argv,ParamFile,SequenceFile,ExonsFile,SRFile);
  
  LengthSequence = analizeFile(SequenceFile);
  sprintf(mess,"DNA sequence file size = %ld bytes",LengthSequence);
  printMess(mess);

  /* 1. Alloc main memory structures */
  printMess("Request Memory to System");
  Sequence   = (char*)        RequestMemorySequence(LengthSequence);
  RSequence  = (char*)        RequestMemorySequence(LengthSequence);
  RequestMemoryGlobalSites(allSites,allSites_r);
  RequestMemoryGlobalExons(allExons,allExons_r);
  RequestMemoryGlobalSortExons(exons,nExons);
  RequestMemoryThreadParams(p);
  RequestMemoryGlobalGenes(genes);
  RequestMemoryGlobalDumpster(dumpster);
  isochores  = (gparam**)      RequestMemoryIsochoresParams();
  dAA        = (dict*)        RequestMemoryAaDictionary();

  if (EVD)
    evidence   = (packEvidence*) RequestMemoryEvidence();
  
  if (SRP)  
    sr         = (packSR*)       RequestMemorySimilarityRegions();

  /* 2. Read prediction parameters file */
  printMess("Reading parameters..."); 
  nIsochores = readparam(ParamFile, isochores);

  /* 3. Reading DNA sequence */
  if ((seqfile = fopen(SequenceFile, "rb"))==NULL) 
    printError("The Sequence file can not be open for read");

  reading = IniReadSequence(seqfile,Locus);
  while (reading!=EOF)
    {
      printMess("Reading DNA sequence");
      reading = ReadSequence(seqfile, Sequence, nextLocus);
      /* Processing sequence: Reverse and complement */
      printMess("Processing DNA sequence");
      LengthSequence = FetchSequence(Sequence, RSequence);
      
      /* Control: length of sequence */
      if (!LengthSequence)
	printError("Sequence empty"); 
      
      /* Verifying length feature */
      if (LengthSequence <= MINLENGTH)
	printError("Sequence too short. Execute Serial geneid on it");  
      
      /* Reading evidence exons */
      if (EVD)
	{
	  printMess("Reading evidence exons");
	  evidence->nvExons = ReadExonsGFF(ExonsFile, evidence, isochores[0]->D);
	  sprintf(mess,"%ld evidence exons read\n", evidence->nvExons);
	  printMess(mess);
	}
      
      /* Reading Similarity to protein regions */
      if (SRP)
	{
	  printMess("Reading Similarity to protein regions");
	  sprintf(mess,"%ld Similarity to protein regions read\n",
		  ReadSR(SRFile, sr, LengthSequence));
	  printMess(mess);
	} 

      /* 4. Parallel processing: sites and exons */
      /* Setting sequence, sites and exon params to threads */
      setThreadParams(p,
		      allSites,allSites_r,
		      allExons,allExons_r,
		      exons,nExons,genes,
		      dumpster);

      /* Threading the predictions */
      CreateThreads(p,threads);

      /* Waiting for the threads */
      WaitThreads(threads);

      /* Join thread genes */
      printMess("Join best genes");

      JoinGenes(p);

      /* Printing predictions */
      POutput(p, Locus, Sequence, isochores[0], dAA);
    }

  /* 4. The End */
  OutputTime();  
  exit(0);
}



